/*
	(c)2002 bero
	under GPL or notify me
*/

#include <stdio.h>	/* size_t */

const static int diff_lookup[16] = {
	1,3,5,7,9,11,13,15,
	-1,-3,-5,-7,-9,-11,-13,-15,
};

const static int index_scale[16] = {
	0x0e6, 0x0e6, 0x0e6, 0x0e6, 0x133, 0x199, 0x200, 0x266,
	0x0e6, 0x0e6, 0x0e6, 0x0e6, 0x133, 0x199, 0x200, 0x266 //same value for speedup
};

static inline int limit(int val,int min,int max)
{
	if (val<min) return min;
	else if (val>max) return max;
	else return val;
}

void pcm16_to_adpcm(unsigned char *dst,const short *src,size_t length,int sign)
{
	int signal,step;
	signal = 0;
	step = 0x7f;

	length = (length+3)/4;
	do {
		int data,val,diff;

		/* hign nibble */
		diff = (*src++^sign) - signal;
		diff = (diff*8)/step;

		val = abs(diff)/2;
		if (val>7) val = 7;
		if (diff<0) val+=8;

		signal += (step*diff_lookup[val])/8;
		signal = limit(signal,-32768,32767);

		step = (step * index_scale[val]) >> 8;
		step = limit(step,0x7f,0x6000);

		data = val<<4;

		/* low nibble */
		diff = (*src++^sign) - signal;
		diff = (diff*8)/step;

		val = (abs(diff))/2;
		if (val>7) val = 7;
		if (diff<0) val+=8;

		signal += (step*diff_lookup[val])/8;
		signal = limit(signal,-32768,32767);

		step = (step * index_scale[val]) >> 8;
		step = limit(step,0x7f,0x6000);

		data |= val;

		*dst++ = data;

	} while(--length);
}

void pcm8_to_adpcm(unsigned char *dst,const signed char *src,size_t length,int sign)
{
	int signal,step;
	signal = 0;
	step = 0x7f;

	length=(length+1)/2;
	do {
		int data,val,diff;

		/* hign nibble */
		diff = ((*src++^sign)<<8) - signal;
		diff = (diff*8)/step;

		val = abs(diff)/2;
		if (val>7) val = 7;
		if (diff<0) val+=8;

		signal += (step*diff_lookup[val])/8;
		signal = limit(signal,-32768,32767);

		step = (step * index_scale[val]) >> 8;
		step = limit(step,0x7f,0x6000);

		data = val<<4;

		/* low nibble */
		diff = ((*src++^sign)<<8) - signal;
		diff = (diff*8)/step;

		val = (abs(diff))/2;
		if (val>7) val = 7;
		if (diff<0) val+=8;

		signal += (step*diff_lookup[val])/8;
		signal = limit(signal,-32768,32767);

		step = (step * index_scale[val]) >> 8;
		step = limit(step,0x7f,0x6000);

		data |= val;

		*dst++ = data;

	} while(--length);
}

void pcm16_to_adpcm_direct(unsigned int addr,const short *src,size_t length,int sign)
{
	int signal,step;
	signal = 0;
	step = 0x7f;

	length = (length+3)/4;
	do {
		int data,val,diff;
		if ((length&7)==0) g2_fifo_wait();

		/* hign nibble */
		diff = (*src++^sign) - signal;
		diff = (diff*8)/step;

		val = abs(diff)/2;
		if (val>7) val = 7;
		if (diff<0) val+=8;

		signal += (step*diff_lookup[val])/8;
		signal = limit(signal,-32768,32767);

		step = (step * index_scale[val]) >> 8;
		step = limit(step,0x7f,0x6000);

		data = val<<4;

		/* low nibble */
		diff = (*src++^sign) - signal;
		diff = (diff*8)/step;

		val = (abs(diff))/2;
		if (val>7) val = 7;
		if (diff<0) val+=8;

		signal += (step*diff_lookup[val])/8;
		signal = limit(signal,-32768,32767);

		step = (step * index_scale[val]) >> 8;
		step = limit(step,0x7f,0x6000);

		data |= val;

		g2_write_8(addr++,data);

	} while(--length);
	g2_fifo_wait();
}

void pcm8_to_adpcm_direct(unsigned int addr,const signed char *src,size_t length,int sign)
{
	int signal,step;
	signal = 0;
	step = 0x7f;

	length=(length+1)/2;
	do {
		int data,val,diff;
		if ((length&7)==0) g2_fifo_wait();

		/* hign nibble */
		diff = ((*src++^sign)<<8) - signal;
		diff = (diff*8)/step;

		val = abs(diff)/2;
		if (val>7) val = 7;
		if (diff<0) val+=8;

		signal += (step*diff_lookup[val])/8;
		signal = limit(signal,-32768,32767);

		step = (step * index_scale[val]) >> 8;
		step = limit(step,0x7f,0x6000);

		data = val<<4;

		/* low nibble */
		diff = ((*src++^sign)<<8) - signal;
		diff = (diff*8)/step;

		val = (abs(diff))/2;
		if (val>7) val = 7;
		if (diff<0) val+=8;

		signal += (step*diff_lookup[val])/8;
		signal = limit(signal,-32768,32767);

		step = (step * index_scale[val]) >> 8;
		step = limit(step,0x7f,0x6000);

		data |= val;

		g2_write_8(addr++,data);

	} while(--length);
	g2_fifo_wait();
}
